
import time

from upcloud_api import CloudManager, Server, Storage, ZONE, login_user_block

from fabric import Connection as SSH_Connection

from yascheduler.clouds import AbstractCloudAPI


class UpCloudAPI(AbstractCloudAPI):

    name = 'upcloud'

    def __init__(self, config):
        super().__init__(max_nodes=config.getint('clouds', 'upcloud_max_nodes', fallback=None))
        self.client = CloudManager(config.get('clouds', 'upcloud_login'), config.get('clouds', 'upcloud_pass'))
        self.client.authenticate()
        self.config = config

    def init_key(self):
        super().init_key()
        self.login_user = login_user_block(
            username=self.ssh_user,
            ssh_keys=[self.public_key],
            create_password=False,
        )

    def create_node(self):
        assert self.ssh_custom_key

        server = self.client.create_server(Server(
            core_number=8,
            memory_amount=4096,
            hostname=self.get_rnd_name('node'),
            zone=ZONE.London,
            storage_devices=[
                Storage(os='Debian 10.0', size=40)
            ],
            login_user=self.login_user
        ))
        ip = server.get_public_ip()
        self._log.info('CREATED %s' % ip)
        self._log.info('WAITING FOR START...')
        time.sleep(30)

        # warm up
        for _ in range(10):
            ssh_conn = SSH_Connection(
                host=ip, user=self.ssh_user, connect_kwargs=self.ssh_custom_key
            )
            try: ssh_conn.run('whoami', hide=True)
            except: time.sleep(5)
            else: break

        return ip

    def delete_node(self, ip):
        for server in self.client.get_servers():
            if server.get_public_ip() == ip:
                server.stop()
                self._log.info('WAITING FOR STOP...')
                time.sleep(20)
                while True:
                    try: server.destroy()
                    except: time.sleep(5)
                    else: break
                for storage in server.storage_devices:
                      storage.destroy()
                self._log.info('DELETED %s' % ip)
                break
        else:
            self._log.info('NODE %s NOT DELETED AS UNKNOWN' % ip)



#!/usr/bin/env python

import json
import logging
import os
import queue
import random
import string
import tempfile
from configparser import ConfigParser
from datetime import datetime, timedelta
from collections import Counter
from typing import List, Optional

from fabric import Connection as SSH_Connection
from yascheduler import connect_db, CONFIG_FILE, SLEEP_INTERVAL, N_IDLE_PASSES
import yascheduler.clouds
from yascheduler.engines import init_engines, get_engines_check_cmd
from yascheduler.time import sleep_until
from yascheduler.webhook_worker import WebhookWorker, WebhookTask

logging.basicConfig(level=logging.INFO)


class Yascheduler(object):

    STATUS_TO_DO = 0
    STATUS_RUNNING = 1
    STATUS_DONE = 2

    _log: logging.Logger
    _webhook_queue: "queue.Queue[WebhookTask]"
    _webhook_threads: List[WebhookWorker]

    def __init__(self, config, logger: Optional[logging.Logger] = None):
        if logger:
            self._log = logger.getChild(self.__class__.__name__)
        else:
            self._log = logging.getLogger(self.__class__.__name__)
        self.config = config
        self.connection, self.cursor = connect_db(config)
        self.ssh_conn_pool = {}
        self.ssh_custom_key = {}
        self.clouds = None
        self.engines = init_engines(config)
        self._webhook_queue = queue.Queue()
        webhook_thread_num = int(
            config.get("local", "webhook_threads", fallback=2)
        )
        self._webhook_threads = []
        for i in range(webhook_thread_num):
            t = WebhookWorker(
                name=f"WebhookThread[{i}]",
                logger=self._log,
                task_queue=self._webhook_queue,
            )
            self._webhook_threads.append(t)

    def start(self) -> None:
        for t in self._webhook_threads:
            t.start()

    def queue_get_resources(self):
        self.cursor.execute('SELECT ip, ncpus, enabled, cloud FROM yascheduler_nodes;')
        return self.cursor.fetchall()

    def queue_get_resource(self, ip):
        self.cursor.execute(
            (
                "SELECT ip, ncpus, enabled, cloud "
                "FROM yascheduler_nodes WHERE ip=(%S);"
            ),
            (ip),
        )
        return self.cursor.fetchone()

    def queue_get_task(self, task_id):
        self.cursor.execute('SELECT label, metadata, ip, status FROM yascheduler_tasks WHERE task_id=%s;' % task_id)
        row = self.cursor.fetchone()
        if not row:
            return None
        return dict(task_id=task_id, label=row[0], metadata=row[1], ip=row[2], status=row[3])

    def queue_get_tasks_to_do(self, num_nodes):
        self.cursor.execute('SELECT task_id, label, metadata FROM yascheduler_tasks WHERE status=%s LIMIT %s;',
                            (self.STATUS_TO_DO, num_nodes))
        return [dict(task_id=row[0], label=row[1], metadata=row[2]) for row in self.cursor.fetchall()]

    def queue_get_tasks(self, jobs=None, status=None):
        if jobs is not None and status is not None:
            raise ValueError("jobs can be selected only by status or by task ids")
        if jobs is None and status is None:
            raise ValueError("jobs can only be selected by status or by task ids")
        if status is not None:
            query_string = 'status IN ({})'.format(', '.join(['%s'] * len(status)))
            params = status
        else:
            query_string = 'task_id IN ({})'.format(', '.join(['%s'] * len(jobs)))
            params = jobs

        sql_statement = 'SELECT task_id, label, ip, status FROM yascheduler_tasks WHERE {};'.format(query_string)
        self.cursor.execute(sql_statement, params)
        return [dict(task_id=row[0], label=row[1], ip=row[2], status=row[3]) for row in self.cursor.fetchall()]

    def enqueue_task_event(self, task_id: int) -> None:
        task = self.queue_get_task(task_id) or {}
        wt = WebhookTask.from_dict(task)
        self._webhook_queue.put(wt)

    def queue_set_task_running(self, task_id, ip):
        self.cursor.execute("UPDATE yascheduler_tasks SET status=%s, ip=%s WHERE task_id=%s;",
                            (self.STATUS_RUNNING, ip, task_id))
        self.connection.commit()
        self.enqueue_task_event(task_id)

    def queue_set_task_done(self, task_id, metadata):
        self.cursor.execute("UPDATE yascheduler_tasks SET status=%s, metadata=%s WHERE task_id=%s;",
                            (self.STATUS_DONE, json.dumps(metadata), task_id))
        self.connection.commit()
        self.enqueue_task_event(task_id)
        #if self.clouds:
        # TODO: free-up CloudAPIManager().tasks

    def queue_submit_task(self, label, metadata, engine):

        if engine not in self.engines:
            raise RuntimeError("Engine %s requested, but not supported" % engine)

        for input_file in self.engines[engine]['input_files']:
            if input_file not in metadata:
                raise RuntimeError("Input file %s was not provided" % input_file)

        metadata['engine'] = engine
        metadata['remote_folder'] = os.path.join(self.config.get('remote', 'data_dir'),
                                                 '_'.join([datetime.now().strftime('%Y%m%d_%H%M%S'),
                                                           ''.join([random.choice(string.ascii_lowercase) for _ in range(4)])]))

        self.cursor.execute("""INSERT INTO yascheduler_tasks (label, metadata, ip, status)
            VALUES ('{label}', '{metadata}', NULL, {status})
            RETURNING task_id;""".format(
            label=label,
            metadata=json.dumps(metadata),
            status=self.STATUS_TO_DO
        ))
        self.connection.commit()
        self._log.info(':::submitted: %s' % label)
        return self.cursor.fetchone()[0]

    def ssh_connect(self, new_nodes):
        old_nodes = self.ssh_conn_pool.keys()

        ip_cloud_map = {}
        resources = self.queue_get_resources()
        for row in resources:
            if row[0] in new_nodes:
                ip_cloud_map[row[0]] = row[3]

        for ip in set(old_nodes) - set(new_nodes):
            self.ssh_conn_pool[ip].close()
            del self.ssh_conn_pool[ip]
        for ip in set(new_nodes) - set(old_nodes):
            cloud = self.clouds and self.clouds.apis.get(ip_cloud_map.get(ip))
            ssh_user = (
                cloud and cloud.ssh_user or self.config.get("remote", "user")
            )
            self.ssh_conn_pool[ip] = SSH_Connection(
                host=ip, user=ssh_user, connect_kwargs=self.ssh_custom_key
            )

        self._log.info(
            "Nodes to watch: %s" % ", ".join(self.ssh_conn_pool.keys())
        )
        if not self.ssh_conn_pool:
            self._log.warning('No nodes set!')

    def ssh_run_task(self, ip, ncpus, label, metadata):
        assert not self.ssh_check_task(ip), \
            "Cannot run the task %s at host %s, as this host is already occupied with another task!" % (
            label, ip) # TODO handle this situation

        assert metadata['remote_folder']
        assert metadata['engine'] in self.engines

        try:
            self.ssh_conn_pool[ip].run('mkdir -p %s' % metadata['remote_folder'], hide=True)
        except Exception as err:
            self._log.error('SSH spawn cmd error: %s' % err)
            return False

        with tempfile.NamedTemporaryFile() as tmp: # NB beware overflown remote
            for input_file in self.engines[metadata['engine']]['input_files']:
                tmp.write(metadata[input_file].encode('utf-8'))
                tmp.flush()
                self.ssh_conn_pool[ip].put(tmp.name, metadata['remote_folder'] + '/' + input_file)
                tmp.seek(0)
                tmp.truncate()

        # placeholders {path} and {ncpus} are supported
        run_cmd = self.engines[metadata['engine']]['spawn'].format(
            path=metadata['remote_folder'],
            ncpus=ncpus or '`grep -c ^processor /proc/cpuinfo`'
        )
        self._log.debug(run_cmd)

        try:
            self.ssh_conn_pool[ip].run(run_cmd, hide=True, disown=True)
        except Exception as err:
            self._log.error('SSH spawn cmd error: %s' % err)
            return False

        return True

    def ssh_check_node(self, ip):
        try:
            node_info = self.queue_get_resource(ip)
            cloud = (
                self.clouds and self.clouds.apis.get(node_info and node_info[3])
            )
            ssh_user = (
                cloud and cloud.ssh_user or self.config.get("remote", "user")
            )
            with SSH_Connection(
                host=ip,
                user=ssh_user,
                connect_kwargs=self.ssh_custom_key,
                connect_timeout=5,
            ) as conn:
                result = str( conn.run(get_engines_check_cmd(self.engines), hide=True) )
                for engine_name in self.engines:
                    if self.engines[engine_name]['run_marker'] in result:
                        self._log.error(
                            "Cannot add a busy with %s resourse %s@%s"
                            % (engine_name, ssh_user, ip)
                        )
                        return False

        except Exception as err:
            self._log.error(
                "Host %s@%s is unreachable due to: %s"
                % (self.config.get("remote", "user"), ip, err)
            )
            return False

        return True

    def ssh_check_task(self, ip):
        assert ip in self.ssh_conn_pool, "Node %s was referred by active task, however absent in node list" % ip
        try:
            result = str( self.ssh_conn_pool[ip].run(get_engines_check_cmd(self.engines), hide=True) )
        except Exception as err:
            self._log.error('SSH status cmd error: %s' % err)
            # TODO handle that situation properly, re-assign ip, etc.
            result = ""

        for engine in self.engines.values():
            if engine['run_marker'] in result:
                return True

        return False

    def ssh_get_task(self, ip, engine, work_folder, store_folder, remove=True):
        for output_file in self.engines[engine]['output_files']:
            try:
                self.ssh_conn_pool[ip].get(work_folder + '/' + output_file, store_folder + '/' + output_file)
            except IOError as err:
                # TODO handle that situation properly
                self._log.error(
                    "Cannot scp %s/%s: %s" % (work_folder, output_file, err)
                )
                if 'Connection timed out' in str(err): break

        if remove:
            self.ssh_conn_pool[ip].run('rm -rf %s' % work_folder, hide=True)

    def clouds_allocate(self, on_task):
        if self.clouds:
            self.clouds.allocate(on_task)

    def clouds_deallocate(self, ips):
        if self.clouds:
            self.clouds.deallocate(ips)

    def clouds_get_capacity(self, resources):
        if self.clouds:
            return self.clouds.get_capacity(resources)
        return 0

    def stop(self):
        self._log.info("Stopping threads...")
        for t in self._webhook_threads:
            t.stop()
            t.join()


def daemonize(log_file=None):
    logger = get_logger(log_file)
    config = ConfigParser()
    config.read(CONFIG_FILE)

    yac, clouds = clouds.yascheduler, yac.clouds = Yascheduler(
        config
    ), yascheduler.clouds.CloudAPIManager(config, logger=logger)
    logging.getLogger('Yascheduler').setLevel(logging.DEBUG)
    clouds.initialize()
    yac.start()

    chilling_nodes = Counter() # ips vs. their occurences

    logger.debug('Available computing engines: %s' % ', '.join([engine_name for engine_name in yac.engines]))

    def step():
        resources = yac.queue_get_resources()
        all_nodes = [item[0] for item in resources if '.' in item[0]] # NB provision nodes have fake ips
        if sorted(yac.ssh_conn_pool.keys()) != sorted(all_nodes):
            yac.ssh_connect(all_nodes)

        enabled_nodes = {item[0]: item[1] for item in resources if item[2]}
        free_nodes = list(enabled_nodes.keys())

        # (I.) Tasks de-allocation clause
        tasks_running = yac.queue_get_tasks(status=(yac.STATUS_RUNNING,))
        logger.debug('running %s tasks: %s' % (len(tasks_running), tasks_running))
        for task in tasks_running:
            if yac.ssh_check_task(task['ip']):
                try: free_nodes.remove(task['ip'])
                except ValueError: pass
            else:
                ready_task = yac.queue_get_task(task['task_id'])
                webhook_url = ready_task['metadata'].get('webhook_url')
                store_folder = ready_task['metadata'].get('local_folder') or \
                    os.path.join(config.get('local', 'data_dir'),
                                 os.path.basename(ready_task['metadata']['remote_folder']))
                os.makedirs(store_folder, exist_ok=True) # TODO OSError if restart or invalid data_dir
                yac.ssh_get_task(ready_task['ip'], ready_task['metadata']['engine'], ready_task['metadata']['remote_folder'], store_folder)
                ready_task['metadata'] = dict(
                    remote_folder=ready_task['metadata']['remote_folder'],
                    local_folder=store_folder,
                )
                if webhook_url:
                    ready_task['metadata']['webhook_url'] = webhook_url
                yac.queue_set_task_done(ready_task['task_id'], ready_task['metadata'])
                logger.info(':::task_id={} {} done and saved in {}'.format(task['task_id'], ready_task['label'],
                                                                ready_task['metadata'].get('local_folder')))
                # TODO here we might want to notify our data consumers in an event-driven manner
                # TODO but how to do it quickly or in the background?

        # (II.) Resourses and tasks allocation clause
        clouds_capacity = yac.clouds_get_capacity(resources)
        if free_nodes or clouds_capacity:
            for task in yac.queue_get_tasks_to_do(clouds_capacity + len(free_nodes)):
                if not free_nodes:
                    yac.clouds_allocate(task['task_id'])
                    continue
                random.shuffle(free_nodes)
                ip = free_nodes.pop()
                logger.info(':::submitting task_id=%s %s to %s' % (task['task_id'], task['label'], ip))

                if yac.ssh_run_task(ip, enabled_nodes[ip], task['label'], task['metadata']):
                    yac.queue_set_task_running(task['task_id'], ip)

        # (III.) Resourses de-allocation clause
        if free_nodes: # candidates for removal
            chilling_nodes.update(free_nodes)
            deallocatable = Counter([
                x[0] for x in filter(lambda x: x[1] >= N_IDLE_PASSES, chilling_nodes.most_common())
            ])
            if deallocatable:
                yac.clouds_deallocate(list(deallocatable.elements()))
                chilling_nodes.subtract(deallocatable)

        # process results of allocators
        clouds.do_async_work()

        # print stats
        nodes = yac.queue_get_resources()
        enabled_nodes = list(filter(lambda x: x[2], nodes))
        logger.info(
            "NODES:\tenabled: %s\ttotal: %s",
            str(len(enabled_nodes)),
            str(len(nodes))
        )
        logger.info(
            "TASKS:\trunning: %s\tto do: %s\tdone: %s",
            len(yac.queue_get_tasks(status=(yac.STATUS_RUNNING,))),
            len(yac.queue_get_tasks(status=(yac.STATUS_TO_DO,))),
            len(yac.queue_get_tasks(status=(yac.STATUS_DONE,))),
        )

    # The main scheduler loop
    try:
        while True:
            end_time = datetime.now() + timedelta(seconds=SLEEP_INTERVAL)
            step()
            sleep_until(end_time)
    except KeyboardInterrupt:
        clouds.stop()
        yac.stop()


def get_logger(log_file):
    logger = logging.getLogger('yascheduler')
    logger.setLevel(logging.DEBUG)

    if log_file:
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


if __name__ == "__main__":
    daemonize()



#!/usr/bin/env python3

import logging
import threading
from datetime import datetime, timedelta
from typing import Optional
from yascheduler.variables import SLEEP_INTERVAL
from yascheduler.time import sleep_until


class BackgroundWorker(threading.Thread):
    _kill: threading.Event
    _log: logging.Logger
    _sleep_interval: float = SLEEP_INTERVAL

    def __init__(
        self,
        logger: Optional[logging.Logger] = None,
        **kwargs,
    ):
        super().__init__(target=self.run, **kwargs)
        if logger:
            self._log = logger.getChild(self.name)
        else:
            self._log = logging.getLogger(self.name)
        self._kill = threading.Event()

    def stop(self):
        self._log.info("Stopping thread...")
        self._kill.set()

    def do_work(self):
        raise NotImplementedError()

    def run(self):
        self._log.info("Thread started")
        while not self._kill.is_set():
            end_time = datetime.now() + timedelta(seconds=self._sleep_interval)
            self.do_work()
            sleep_until(end_time)
        self._log.info("Thread stopped")




from .variables import (
    __version__,
    CONFIG_FILE, LOG_FILE, PID_FILE,
    SLEEP_INTERVAL, N_IDLE_PASSES, DEFAULT_NODES_PER_PROVIDER
)


def connect_db(config):
    import pg8000

    connection = pg8000.connect(
        user=config.get('db', 'user'),
        password=config.get('db', 'password'),
        database=config.get('db', 'database'),
        host=config.get('db', 'host'),
        port=config.getint('db', 'port')
    )
    cursor = connection.cursor()
    return connection, cursor


def has_node(config, ip):
    connection, cursor = connect_db(config)
    cursor.execute('SELECT * FROM yascheduler_nodes WHERE ip=%s;', [ip])
    result = cursor.fetchall()
    connection.close()
    return result


def add_node(config, ip, ncpus=None, cloud=None, provisioned=False):
    connection, cursor = connect_db(config)
    if cloud and provisioned:
        cursor.execute("""DELETE FROM yascheduler_nodes WHERE ip IN (
            SELECT ip FROM yascheduler_nodes WHERE ip LIKE 'prov' || '%%' AND cloud='{}' LIMIT 1
        );""".format(cloud))
    cursor.execute('INSERT INTO yascheduler_nodes (ip, ncpus, cloud) VALUES (%s, %s, %s);', [ip, ncpus, cloud])
    connection.commit()
    connection.close()
    return True


def remove_node(config, ip):
    connection, cursor = connect_db(config)
    cursor.execute('DELETE FROM yascheduler_nodes WHERE ip=%s;', [ip])
    connection.commit()
    connection.close()
    return True




import warnings


def init_engines(config):

    engines = {}

    for section in config.sections():
        if not section.startswith('engine.'):
            continue

        name = section[7:]

        engines[name] = dict(config[section])

        assert 'spawn' in engines[name]
        assert 'check' in engines[name]
        assert 'run_marker' in engines[name]
        assert engines[name]['spawn'].startswith('nohup')
        assert 'input_files' in engines[name] and engines[name]['input_files']
        engines[name]['input_files'] = [x.strip() for x in filter(None, engines[name]['input_files'].split())]
        assert 'output_files' in engines[name] and engines[name]['output_files']
        engines[name]['output_files'] = [x.strip() for x in filter(None, engines[name]['output_files'].split())]

        # TODO own sleep_interval

        if 'deployable' not in engines[name] or not len(engines[name]['deployable'].strip()):
            warnings.warn('Engine %s has no *deployable* config set, cloud usage is impossible' % name)

    if not engines:
        raise RuntimeError('No engines were set up')

    return engines


def get_engines_check_cmd(engines):

    cmd = []

    for name in engines:
        cmd.append(engines[name]['check'].strip())

    return ' && '.join(cmd)


#!/usr/bin/env python3

import inspect
import queue
import dataclasses
from dataclasses import dataclass
from typing import Any, Dict, Optional
import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
from yascheduler.background_worker import BackgroundWorker


def from_dict_to_dataclass(cls, data: Dict[str, Any]):
    new_dict = {
        key: (
            data.get(key)
            if val.default == val.empty
            else data.get(key, val.default)
        )
        for key, val in inspect.signature(cls).parameters.items()
    }
    return cls(**new_dict)


@dataclass
class WebhookTaskMetadata:
    webhook_url: Optional[str]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "WebhookTaskMetadata":
        return from_dict_to_dataclass(cls, data)


@dataclass
class WebhookTask:
    status: int
    metadata: WebhookTaskMetadata

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "WebhookTask":
        return from_dict_to_dataclass(cls, data)

    def __post_init__(self) -> None:
        if isinstance(self.metadata, dict):
            self.metadata = WebhookTaskMetadata.from_dict(self.metadata)


@dataclass
class WebhookPayload:
    status: int


class WebhookWorker(BackgroundWorker):
    _task_queue: "queue.Queue[WebhookTask]"
    http: requests.Session

    def __init__(
        self,
        task_queue: "queue.Queue[WebhookTask]",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._task_queue = task_queue
        retry_strategy = Retry(total=5, backoff_factor=1)
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.http = requests.Session()
        self.http.mount("https://", adapter)
        self.http.mount("http://", adapter)

    def do_work(self) -> None:
        try:
            t = self._task_queue.get(False)
        except queue.Empty:
            return

        if not isinstance(t.metadata.webhook_url, str):
            self._task_queue.task_done()
            return

        self._log.info(f"Executing webhook to {t.metadata.webhook_url}")
        payload = WebhookPayload(status=t.status)
        try:
            response = self.http.post(
                url=t.metadata.webhook_url,
                json=dataclasses.asdict(payload),
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            self._log.info(
                f"Webhook to {t.metadata.webhook_url} failed: {str(e)}"
            )

        self._task_queue.task_done()



"""
Console scripts for yascheduler
"""
import os
import argparse
from configparser import ConfigParser

from pg8000 import ProgrammingError
from fabric import Connection as SSH_Connection
from paramiko.rsakey import RSAKey
from invoke.exceptions import UnexpectedExit

from yascheduler import has_node, add_node, remove_node
from yascheduler.variables import CONFIG_FILE
from yascheduler.scheduler import Yascheduler


def submit():
    parser = argparse.ArgumentParser(description="Submit task to yascheduler daemon")
    parser.add_argument('script')

    args = parser.parse_args()
    if not os.path.isfile(args.script):
        raise ValueError("Script parameter is not a file name")

    inputs = {}
    with open(args.script) as f:
        for l in f.readlines():
            try:
                k, v = l.split('=')
                inputs[k.strip()] = v.strip()
            except ValueError:
                pass
    config = ConfigParser()
    config.read(CONFIG_FILE)
    yac = Yascheduler(config)
    task_id = yac.queue_submit_task(inputs['LABEL'],
                                    {'structure': open(inputs['STRUCT']).read(),
                                     'input': open(inputs['INPUT']).read(),
                                     'local_folder': os.getcwd()}) # TODO

    print("Successfully submitted task: {}".format(task_id))
    yac.connection.close()


def check_status():
    parser = argparse.ArgumentParser(description="Submit task to yascheduler daemon")
    parser.add_argument('-j', '--jobs', required=False, default=None, nargs='*')
    parser.add_argument('-v', '--view', required=False, default=None, nargs='?', type=bool, const=True)
    parser.add_argument('-o', '--convergence', required=False, default=None, nargs='?', type=bool, const=True, help='needs -v option')
    parser.add_argument('-i', '--info', required=False, default=None, nargs='?', type=bool, const=True)
    #parser.add_argument('-k', '--kill', required=False, default=None, nargs='?', type=bool, const=True)

    args = parser.parse_args()
    config = ConfigParser()
    config.read(CONFIG_FILE)
    yac = Yascheduler(config)
    statuses = {
        yac.STATUS_TO_DO: "QUEUED",
        yac.STATUS_RUNNING: "RUNNING",
        yac.STATUS_DONE: "FINISHED"
    }
    local_parsing_ready, local_calc_snippet = False, False

    if args.jobs:
        tasks = yac.queue_get_tasks(jobs=args.jobs)
    else:
        tasks = yac.queue_get_tasks(status=(yac.STATUS_RUNNING, yac.STATUS_TO_DO))

    if args.view: # or args.kill:
        if not tasks:
            print('NO MATCHING TASKS FOUND')
            return
        ssh_custom_key = {}
        for filename in os.listdir(config.get('local', 'data_dir')):
            if not filename.startswith('yakey') or not os.path.isfile(
                os.path.join(config.get('local', 'data_dir'), filename)):
                continue
            key_path = os.path.join(config.get('local', 'data_dir'), filename)
            pmk_key = RSAKey.from_private_key_file(key_path)
            print('LOADED KEY %s' % key_path)
            ssh_custom_key = {'pkey': pmk_key}
            break

    if args.convergence:
        try:
            from pycrystal import CRYSTOUT, CRYSTOUT_Error
            from numpy import nan
            local_parsing_ready = True
        except: pass

    if args.view:
        yac.cursor.execute(
            (
                "SELECT t.task_id, t.label, t.metadata, t.ip, n.cloud "
                "FROM yascheduler_tasks AS t "
                "JOIN yascheduler_nodes AS n ON n.ip=t.ip "
                "WHERE status=%s AND task_id IN (%s);"
            ),
            (
                yac.STATUS_RUNNING,
                ", ".join([str(task["task_id"]) for task in tasks]),
            ),
        )
        for row in yac.cursor.fetchall():
            ssh_user = config.get(
                "clouds", f"{row[4]}", fallback=config.get("remote", "user")
            )
            print(
                "." * 50
                + "ID%s %s at %s@%s:%s"
                % (row[0], row[1], ssh_user, row[3], row[2]["remote_folder"])
            )
            ssh_conn = SSH_Connection(
                host=row[3], user=ssh_user, connect_kwargs=ssh_custom_key
            )
            try:
                result = ssh_conn.run('tail -n15 %s/OUTPUT' % row[2]['remote_folder'], hide=True)
            except UnexpectedExit:
                print('OUTDATED TASK, SKIPPING')
            else:
                print(result.stdout)

            if local_parsing_ready:
                local_calc_snippet = os.path.join(config.get('local', 'data_dir'), 'local_calc_snippet.tmp')
                try:
                    ssh_conn.get(row[2]['remote_folder'] + '/OUTPUT', local_calc_snippet)
                except IOError as err:
                    continue
                try:
                    calc = CRYSTOUT(local_calc_snippet)
                except CRYSTOUT_Error as err:
                    print(err)
                    continue
                output_lines = ''
                if calc.info['convergence']:
                    output_lines += str(calc.info['convergence']) + "\n"
                if calc.info['optgeom']:
                    for n in range(len(calc.info['optgeom'])):
                        try:
                            ncycles = calc.info['ncycles'][n]
                        except IndexError:
                            ncycles = "^"
                        output_lines += "{:8f}".format(calc.info['optgeom'][n][0] or nan) + "  " + \
                                        "{:8f}".format(calc.info['optgeom'][n][1] or nan) + "  " + \
                                        "{:8f}".format(calc.info['optgeom'][n][2] or nan) + "  " + \
                                        "{:8f}".format(calc.info['optgeom'][n][3] or nan) + "  " + \
                                        "E={:12f}".format(calc.info['optgeom'][n][4] or nan) + " eV" + "  " + \
                                        "(%s)" % ncycles + "\n"
                print(output_lines)

    #elif args.kill:
    #    if not args.jobs:
    #        print('NO JOBS GIVEN')
    #        return
    #    yac.cursor.execute(
    #        'SELECT ip FROM yascheduler_tasks WHERE status=%s AND task_id IN (%s);' % (
    #        yac.STATUS_RUNNING, ', '.join([str(task['task_id']) for task in tasks])
    #    ))
    #    for row in yac.cursor.fetchall():
    #        ssh_conn = SSH_Connection(host=row[0], user=config.get('remote', 'user'),
    #            connect_kwargs=ssh_custom_key)
    #        try:
    #            result = ssh_conn.run('pkill %s' % yac.RUNNING_MARKER, hide=True)
    #        except: pass

    elif args.info:
        for task in tasks:
            print('task_id={}\tstatus={}\tlabel={}\tip={}'.format(
                task['task_id'], statuses[task['status']], task['label'], task['ip'] or '-'
            ))

    else:
        for task in tasks:
            print('{}   {}'.format(task['task_id'], statuses[task['status']]))

    yac.connection.close()

    if local_calc_snippet and os.path.exists(local_calc_snippet):
        os.unlink(local_calc_snippet)


def init():
    # service initialization
    install_path = os.path.dirname(__file__)
    # check for systemd (exit status is 0 if there is a process)
    has_systemd = not os.system("pidof systemd")
    if has_systemd:
        _init_systemd(install_path) # NB. will be absent in *service --status-all*
    else:
        _init_sysv(install_path)
    _init_db(install_path)


def _init_systemd(install_path):
    print("Installing systemd service")
    # create unit file in /lib/systemd/system
    src_unit_file = os.path.join(install_path, 'data/yascheduler.service')
    unit_file = os.path.join('/lib/systemd/system/yascheduler.service')
    if not os.path.isfile(unit_file):
        daemon_file = os.path.join(install_path, 'daemon_systemd.py')
        systemd_script = open(src_unit_file).read().replace('%YASCHEDULER_DAEMON_FILE%', daemon_file)
        with open(unit_file, 'w') as f:
            f.write(systemd_script)


def _init_sysv(install_path):

    print("Installing SysV service")

    # create sysv script in /etc/init.d
    src_startup_file = os.path.join(install_path, 'data/yascheduler.sh')
    startup_file = os.path.join('/etc/init.d/yascheduler')
    if not os.path.isfile(startup_file):
        if not os.access(startup_file, os.W_OK):
            print("Error: cannot write to %s" % startup_file)
            return

        daemon_file = os.path.join(install_path, 'daemon_sysv.py')
        sysv_script = open(src_startup_file).read().replace('%YASCHEDULER_DAEMON_FILE%', daemon_file)
        with open(startup_file, 'w') as f:
            f.write(sysv_script)
        # make script executable
        os.chmod(startup_file, 0o755)


def _init_db(install_path):
    # database initialization
    config = ConfigParser()
    config.read(CONFIG_FILE)
    yac = Yascheduler(config)
    schema = open(os.path.join(install_path, 'data', 'schema.sql')).read()
    try:
        for line in schema.split(';'):
            if not line:
                continue
            yac.cursor.execute(line)
        yac.connection.commit()
    except ProgrammingError as e:
        if "already exists" in str(e.args[0]):
            print("Database already initialized!")
        raise


def show_nodes():
    config = ConfigParser()
    config.read(CONFIG_FILE)
    yac = Yascheduler(config)

    yac.cursor.execute('SELECT ip, label, task_id FROM yascheduler_tasks WHERE status=%s;', [yac.STATUS_RUNNING])
    tasks_running = {row[0]: [row[1], row[2]] for row in yac.cursor.fetchall()}

    yac.cursor.execute('SELECT ip, ncpus, enabled, cloud from yascheduler_nodes;')
    for item in yac.cursor.fetchall():
        print("ip=%s ncpus=%s enabled=%s occupied_by=%s (task_id=%s) %s" % tuple(
            [item[0], item[1] or 'MAX', item[2]] + tasks_running.get(item[0], ['-', '-']) + [item[3] or '']
        ))


def manage_node():
    parser = argparse.ArgumentParser(description="Add nodes to yascheduler daemon")
    parser.add_argument('host',
        help='IP[~ncpus]')
    parser.add_argument('--remove-soft', required=False, default=None, nargs='?', type=bool, const=True,
        help='Remove IP delayed')
    parser.add_argument('--remove-hard', required=False, default=None, nargs='?', type=bool, const=True,
        help='Remove IP immediate')

    args = parser.parse_args()
    config = ConfigParser()
    config.read(CONFIG_FILE)

    ncpus = None
    if '~' in args.host:
        args.host, ncpus = args.host.split('~')
        ncpus = int(ncpus)

    yac = Yascheduler(config)

    already_there = has_node(config, args.host)
    if already_there and not args.remove_hard and not args.remove_soft:
        print('Host already in DB: {}'.format(args.host))
        return False

    if not already_there and (args.remove_hard or args.remove_soft):
        print('Host NOT in DB: {}'.format(args.host))
        return False

    if args.remove_hard:
        yac.cursor.execute('SELECT task_id from yascheduler_tasks WHERE ip=%s AND status=%s;', [args.host, yac.STATUS_RUNNING])
        result = yac.cursor.fetchall() or []
        for item in result: # only one item is expected, but here we also account inconsistency case
            yac.cursor.execute('UPDATE yascheduler_tasks SET status=%s WHERE task_id=%s;', [yac.STATUS_DONE, item[0]])
            yac.connection.commit()
            print('An associated task %s at %s is now marked done!' % (item[0], args.host))

        remove_node(config, args.host)
        print('Removed host from yascheduler: {}'.format(args.host))
        return True

    elif args.remove_soft:
        yac.cursor.execute('SELECT task_id from yascheduler_tasks WHERE ip=%s AND status=%s;', [args.host, yac.STATUS_RUNNING])
        if yac.cursor.fetchall():
            print('A task associated, prevent from assigning the new tasks')
            yac.cursor.execute('UPDATE yascheduler_nodes SET enabled=FALSE WHERE ip=%s;', [args.host])
            yac.connection.commit()
            print('Prevented from assigning the new tasks: {}'.format(args.host))
            return True

        else:
            print('No tasks associated, remove node immediately')
            remove_node(config, args.host)
            print('Removed host from yascheduler: {}'.format(args.host))
            return True

    if not yac.ssh_check_node(args.host) or not add_node(config, args.host, ncpus):
        print('Failed to add host to yascheduler: {}'.format(args.host))
        return False

    print('Added host to yascheduler: {}'.format(args.host))
    return True



#!/usr/bin/env python3

import queue
from configparser import ConfigParser
from dataclasses import dataclass
from typing import Dict, List, Optional
from .abstract_cloud_api import AbstractCloudAPI, load_cloudapi
from .. import SLEEP_INTERVAL
from ..background_worker import BackgroundWorker


class CloudWorker(BackgroundWorker):
    _cfg: ConfigParser
    _apis: Dict[str, AbstractCloudAPI]

    def __init__(
        self,
        config: ConfigParser,
        use_apis: List[str],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._cfg = config

        self._apis = {}
        for api_name in use_apis:
            self._apis[api_name] = load_cloudapi(api_name)(config)
            self._apis[api_name]._log = self._log.getChild(api_name)


@dataclass
class AllocateTask:
    api_name: str
    tmp_ip: str


@dataclass
class AllocateResult:
    api_name: str
    tmp_ip: str
    ip: Optional[str] = None
    provisioned: bool = False
    ncpus: Optional[int] = None


class AllocatorWorker(CloudWorker):
    _task_queue: "queue.Queue[AllocateTask]"
    _result_queue: "queue.Queue[AllocateResult]"

    def __init__(
        self,
        task_queue: "queue.Queue[AllocateTask]",
        result_queue: "queue.Queue[AllocateResult]",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._task_queue = task_queue
        self._result_queue = result_queue

    def do_work(self):
        try:
            t = self._task_queue.get(False)
        except queue.Empty:
            return

        r = AllocateResult(api_name=t.api_name, tmp_ip=t.tmp_ip)

        api = self._apis.get(t.api_name)
        if not api:
            self._log.error(f"Unknown cloud API {t.api_name}")
            self._task_queue.task_done()
            return

        try:
            api.init_key()
            self._log.info("Creating a node...")
            r.ip = api.create_node()
            self._log.info(f"Created: {r.ip}")
            api.setup_node(r.ip)
            self._log.info(f"Provisioned: {r.ip}")
            r.provisioned = True
            self._sleep_interval = SLEEP_INTERVAL
        except Exception as e:
            self._log.error(f"Allocation of {t.tmp_ip} failed: {str(e)}")
            # backoff
            self._sleep_interval = min(self._sleep_interval * 1.3, 60)

        self._result_queue.put(r)
        self._task_queue.task_done()


@dataclass
class DeallocateTask:
    api_name: str
    ip: str


@dataclass
class DeallocateResult:
    ip: str


class DeallocatorWorker(CloudWorker):
    _task_queue: "queue.Queue[DeallocateTask]"
    _result_queue: "queue.Queue[DeallocateResult]"

    def __init__(
        self,
        task_queue: "queue.Queue[DeallocateTask]",
        result_queue: "queue.Queue[DeallocateResult]",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._task_queue = task_queue
        self._result_queue = result_queue

    def do_work(self):
        try:
            t = self._task_queue.get(False)
        except queue.Empty:
            return

        r = DeallocateResult(ip=t.ip)

        api = self._apis.get(t.api_name)
        if not api:
            self._log.error(f"Unknown cloud API {t.api_name}")
            self._task_queue.task_done()
            return

        try:
            api.init_key()
            self._log.info(f"Deleting the {r.ip} node...")
            api.delete_node(r.ip)
            self._log.info(f"Node {r.ip} is deleted")
            self._sleep_interval = SLEEP_INTERVAL
        except Exception as e:
            self._log.error(f"Deallocation of {r.ip} failed: {str(e)}")
            # backoff
            self._sleep_interval = min(self._sleep_interval * 1.3, 60)

        self._result_queue.put(r)
        self._task_queue.task_done()



"""
A setuptools script for yet another scheduler
"""

import atexit
import os
import sys
import shutil
from setuptools import setup, find_packages
from setuptools.command.install import install
from yascheduler import __version__, CONFIG_FILE

package_name = "yascheduler"
package_version = __version__


class CustomInstall(install):
    def run(self):
        def _post_install():
            def find_module_path():
                for p in sys.path:
                    if os.path.isdir(p) and package_name in os.listdir(p):
                        return os.path.join(p, package_name)
            install_path = find_module_path()
            src_config = os.path.join(install_path, 'data/yascheduler.conf')
            # create config file in /etc if absent
            if not os.path.isfile(CONFIG_FILE):
                config_dir = os.path.dirname(CONFIG_FILE)
                os.makedirs(config_dir)
                shutil.copy(src_config, CONFIG_FILE)

            # link executable
            try: os.symlink(os.path.join(install_path, 'scheduler.py'), '/usr/bin/yascheduler')
            except Exception: pass

        atexit.register(_post_install)
        install.run(self)


if __name__ == '__main__':

    with open('requirements.txt') as f:
        requirements = f.read().splitlines()

    setup(
        name=package_name,
        version=package_version,
        author="Evgeny Blokhin",
        author_email="eb@tilde.pro",
        description="Yet another computing scheduler and cloud orchestration engine",
        long_description="*Yascheduler* is a simple job scheduler designed for submitting scientific simulations and copying back their results from the computing clouds.",
        license="MIT",
        packages=find_packages(),
        package_data={'yascheduler': ['data/*']},
        install_requires=requirements,
        entry_points={
            "console_scripts": [
                "yasubmit = yascheduler.utils:submit",
                "yastatus = yascheduler.utils:check_status",
                "yanodes = yascheduler.utils:show_nodes",
                "yasetnode = yascheduler.utils:manage_node",
                "yainit = yascheduler.utils:init"
            ],
            "aiida.schedulers": [
                "yascheduler = yascheduler.aiida_plugin:YaScheduler",
            ],
        },
        cmdclass={'install': CustomInstall},
        python_requires='>=3.5',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Physics',
            'Topic :: Scientific/Engineering :: Information Analysis',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'
        ]
    )



#!/usr/bin/env python

import logging
import time
from configparser import ConfigParser
from yascheduler import CONFIG_FILE
from yascheduler.clouds import CloudAPIManager
from yascheduler.scheduler import Yascheduler

logging.basicConfig(level=logging.INFO)

config = ConfigParser()
config.read(CONFIG_FILE)
config.set("local", "allocator_threads", "1")
config.set("local", "deallocator_threads", "0")

yac = Yascheduler(config)
clouds = CloudAPIManager(config)
clouds.yascheduler = yac
clouds.initialize()
tmp_ip = clouds.allocate_node()
time.sleep(10)
clouds.stop()



#!/usr/bin/env python3
"""
Azure cloud provider implementation.

Azure setup:
- Create a dedicated Resource Group:
  `az group create --location "{location}" --resource-group "{rg_name}"`
- Create a dedicated application for yascheduler service
- Assign roles "Network Contributor" and "Virtual Machine Contributor"
  to the app in the created Resource Group:
  ```
  az role assignment create \
    --assignee "{client_id}" \
    --role "{role_name}" \
    --resource-group "{rg_name}
  ```

Configuration in `yascheduler.conf`:
```
[clouds]
; required:
az_tenant_id = ...
az_client_id = ...
az_client_secret = ...
az_subscription_id = ...
az_resource_group = ...
; optional
az_max_nodes = ...
az_user = ...
; optional inputs for infra deployment (example):
az_infra_param_location = westeurope
; optinal imputs for VM deployment (example):
az_vm_param_location = westeurope
```
"""

import json
import logging
import random
import string
from pathlib import Path
from threading import Lock
from typing import cast, Any, Dict, List, NamedTuple, Optional, Union

from configparser import ConfigParser
from azure.identity import ClientSecretCredential
from azure.mgmt.compute.v2021_07_01 import ComputeManagementClient
from azure.mgmt.compute.v2021_07_01.operations import VirtualMachinesOperations
from azure.mgmt.network.v2020_06_01 import NetworkManagementClient
from azure.mgmt.network.v2020_06_01.models import PublicIPAddress
from azure.mgmt.network.v2020_06_01.operations import (
    PublicIPAddressesOperations,
    NetworkInterfacesOperations,
)
from azure.mgmt.resource.resources.v2021_04_01 import ResourceManagementClient
from azure.mgmt.resource.resources.v2021_04_01.models import (
    Dependency,
    Deployment,
    DeploymentExtended,
    DeploymentMode,
    DeploymentProperties,
    ResourceGroup,
)
from azure.mgmt.resource.resources.v2021_04_01.operations import (
    DeploymentsOperations,
)
from azure.core.exceptions import HttpResponseError

from yascheduler.clouds import AbstractCloudAPI

# Azure SDK is too noisy
for logger_name in [
    "azure.core.pipeline.policies.http_logging_policy",
    "azure.identity._internal.get_token_mixin",
]:
    logging.getLogger(logger_name).setLevel(logging.ERROR)

infra_deployment_lock = Lock()

RG_LOCATION_MISMATCH_TMPL = (
    "Resource Group '{}' location mismatch: got '{}' instead of '{}'"
)

DeleteRequestsOperations = Union[
    DeploymentsOperations,
    NetworkInterfacesOperations,
    PublicIPAddressesOperations,
    VirtualMachinesOperations,
]


class DeleteRequest(NamedTuple):
    order: int
    operations: DeleteRequestsOperations
    name: str


class AzureRBACError(Exception):
    operation: str = "read"
    resource_type: Optional[str] = None

    def __init__(self, name: str):
        fstr = "Can't {} {} '{}' - access denied. Please, setup RBAC."
        msg = fstr.format(
            self.operation, self.resource_type or "resource", name
        )
        super().__init__(msg)


class AzureRGReadRBACError(AzureRBACError):
    operation = "read"
    resource_type = "Resource Group"


class AzurePubIPReadRBACError(AzureRBACError):
    operation = "read"
    resource_type = "Public IP Address"


class AzureDeploymentCreateRBACError(AzureRBACError):
    operation = "create"
    resuource_type = "Deployment"


class AzureNotFoundError(Exception):
    resource_type: Optional[str] = None

    def __init__(self, name: str):
        fstr = "{} '{}' - not found."
        msg = fstr.format(self.resource_type or "Resource", name)
        super().__init__(msg)


class AzureRGNotFoundError(AzureNotFoundError):
    resource_type = "Resource Group"


class AzurePubIPNotFoundError(AzureNotFoundError):
    resource_type = "Public IP Address"


class AzureCreateError(Exception):
    resource_type: Optional[str] = None

    def __init__(self, name: str):
        fstr = "{} '{}' failed to create."
        msg = fstr.format(self.resource_type or "Resource", name)
        super().__init__(msg)


class AzureDeploymentCreateError(AzureCreateError):
    resource_type = "Deployment"


class AzureCreatedVMPublicIPNotFoundError(Exception):
    def __init__(self):
        super().__init__("VM created without IP in outputs")


class AzureAPI(AbstractCloudAPI):

    name = "az"
    client_id: str
    location: str
    rg_name: str
    resource_client: ResourceManagementClient
    network_client: NetworkManagementClient
    compute_client: ComputeManagementClient
    infra_tmpl_path: Path
    infra_deployment_name_tmpl: str = "{}-infra-deployment"
    vm_tmpl_path: Path
    vm_deployment_name_tmpl: str = "{}-vm-{}-deployment"

    def __init__(self, config: ConfigParser):
        super().__init__(
            max_nodes=config.getint("clouds", "az_max_nodes", fallback=None)
        )
        self.config = config
        self.client_id = config.get("clouds", "az_client_id")
        self.location = config.get(
            "clouds", "az_location", fallback="westeurope"
        )
        self.rg_name = config.get(
            "clouds", "az_resource_group_name", fallback="YaScheduler-VM-rg"
        )
        self.infra_tmpl_path = Path(
            config.get(
                "clouds",
                "az_infra_tmpl_path",
                fallback=Path(__file__).parent.absolute()
                / Path("azure_infra_tmpl.json"),
            )
        )
        self.vm_tmpl_path = Path(
            config.get(
                "clouds",
                "az_vm_tmpl_path",
                fallback=Path(__file__).parent.absolute()
                / Path("azure_vm_tmpl.json"),
            )
        )
        credential = ClientSecretCredential(
            tenant_id=config.get("clouds", "az_tenant_id"),
            client_id=self.client_id,
            client_secret=config.get("clouds", "az_client_secret"),
        )
        subscription_id = config.get("clouds", "az_subscription_id")
        self.resource_client = ResourceManagementClient(
            credential,
            subscription_id,
            api_version="2021-04-01",
        )
        self.network_client = NetworkManagementClient(
            credential, subscription_id
        )
        self.compute_client = ComputeManagementClient(
            credential, subscription_id
        )

    @property
    def ssh_user(self) -> str:
        "Default SSH user for azure"
        ssh_user = super().ssh_user
        if ssh_user == "root":
            self._log.warn("Root user not supported on Azure")
            ssh_user = "yascheduler"
        return ssh_user

    @property
    def cloud_config_data(self) -> Dict[str, Any]:
        "cloud-config for azure"
        my_boot_cmds = [
            # see https://github.com/MicrosoftDocs/azure-docs/issues/82500
            "systemctl mask waagent-apt.service",
        ]
        data = super().cloud_config_data
        data["bootcmd"] = my_boot_cmds + data.get("bootcmd", [])
        return data

    def _get_conf_by_prefix(self, section: str, prefix: str) -> Dict[str, str]:
        "Get part of config by section and prefix as dict"
        filtered = filter(
            lambda x: x[0].startswith(prefix), self.config.items(section)
        )
        prefix_removed = map(
            lambda x: (
                x[0][len(prefix):] if x[0].startswith(prefix) else x[0],
                x[1],
            ),
            filtered,
        )
        return dict(prefix_removed)

    def get_rg(self) -> ResourceGroup:
        "Check Resource Group"
        try:
            rg_result = self.resource_client.resource_groups.get(self.rg_name)
            if self.location != rg_result.location:
                msg = RG_LOCATION_MISMATCH_TMPL.format(
                    self.rg_name, rg_result.location, self.location
                )
                self._log.warning(msg)
            return rg_result
        except HttpResponseError as e:
            code = getattr(e, "error", None) and getattr(e.error, "code", None)
            if code == "AuthorizationFailed":
                raise AzureRGReadRBACError(self.rg_name) from e
            if code == "ResourceGroupNotFound":
                raise AzureRGNotFoundError(self.rg_name) from e
            raise e

    def get_pip(self, name: str) -> PublicIPAddress:
        "Get Public IP Address by name"
        try:
            return self.network_client.public_ip_addresses.get(
                self.rg_name, name
            )
        except HttpResponseError as e:
            code = getattr(e, "error", None) and getattr(e.error, "code", None)
            if code == "AuthorizationFailed":
                raise AzurePubIPReadRBACError(name) from e
            raise AzurePubIPNotFoundError(name) from e

    def create_deployment(
        self, name: str, tmpl: object, params: Dict[str, Any]
    ) -> DeploymentExtended:
        "Create deployment"
        properties = DeploymentProperties(
            mode=DeploymentMode.incremental,
            template=tmpl,
            parameters={k: {"value": v} for k, v in params.items()},
        )
        try:
            res = self.resource_client.deployments.begin_create_or_update(
                resource_group_name=self.rg_name,
                deployment_name=name,
                parameters=Deployment(properties=properties),
            ).result()
            self._log.info(f"Deployment {name} created/updated")
            return res
        except HttpResponseError as e:
            code = getattr(e, "error", None) and getattr(e.error, "code", None)
            if code == "AuthorizationFailed":
                raise AzureDeploymentCreateRBACError(name) from e
            self._log.error(e)
            raise AzureDeploymentCreateError(name) from e

    def create_infra_deployment(self) -> Dict[str, Any]:
        "Create deployment with common infrastructure parts"
        params = self._get_conf_by_prefix("clouds", "az_infra_param_")
        name = self.infra_deployment_name_tmpl.format(self.rg_name)
        with open(self.infra_tmpl_path, "r") as fd:
            tmpl = json.load(fd)
        res = self.create_deployment(name, tmpl, params)
        return res.properties and res.properties.outputs or {}

    def create_vm_deployment(self, infra_outputs) -> Dict[str, Any]:
        "Create deployment with VM parts"
        rnd_id = "".join(
            [random.choice(string.ascii_lowercase) for _ in range(8)]
        )
        name = self.vm_deployment_name_tmpl.format(self.rg_name, rnd_id)
        params = {
            "namePrefix": rnd_id,
            "adminPublicKey": self.public_key,
            "customData": "#cloud-config\n"
            + json.dumps(self.cloud_config_data),
        }
        # inherit params from infra deployment outputs
        inherit_infra_params = [
            "projectName",
            "location",
            "networkSecurityGroupName",
            "virtualNetworkName",
            "subnetName",
        ]
        for k, v in infra_outputs.items():
            if k in inherit_infra_params and type(v) == dict:
                params[k] = v.get("value")
        # load from config
        params.update(self._get_conf_by_prefix("clouds", "az_vm_param_"))

        with open(self.vm_tmpl_path, "r") as fd:
            tmpl = json.load(fd)
        res = self.create_deployment(name, tmpl, params)
        return res.properties and res.properties.outputs or {}

    def create_node(self):
        infra_deployment_lock.acquire()
        try:
            infra_outputs = self.create_infra_deployment()
        finally:
            infra_deployment_lock.release()
        vm_outputs = self.create_vm_deployment(infra_outputs)

        ip_name: Optional[str] = vm_outputs.get("publicIpAddressName", {}).get(
            "value"
        )
        if not ip_name:
            raise AzureCreatedVMPublicIPNotFoundError()
        ip_address = self.get_pip(ip_name).ip_address
        if not ip_address:
            raise AzureCreatedVMPublicIPNotFoundError()

        # wait node up and ready
        self._run_ssh_cmd_with_backoff(
            ip_address, cmd="cloud-init status --wait", max_time=600
        )

        return ip_address

    def _run_del_reqs(self, reqs: List[DeleteRequest]) -> None:
        for req in sorted(reqs, key=lambda x: x[0]):
            self._log.info(f"Removing {req.name}...")
            try:
                req.operations.begin_delete(self.rg_name, req.name).result()
                self._log.info(f"{req.name} removed")
            except Exception as e:
                self._log.info(f"Can't remove {req.name}: {str(e)}")

    def delete_node(self, ip):
        del_reqs: List[DeleteRequest] = []

        # find Public IP Address
        pip_obj = None
        for i in self.network_client.public_ip_addresses.list(self.rg_name):
            i = cast(PublicIPAddress, i)
            if i.ip_address != ip:
                continue
            pip_obj = i
            req = DeleteRequest(
                99, self.network_client.public_ip_addresses, cast(str, i.name)
            )
            del_reqs.append(req)

        if not pip_obj:
            self._log.error(f"Public IP {ip} not found")
            return
        pip_tags = cast(Dict[str, str], pip_obj.tags) or {}

        # find Deployment
        deployment_id = pip_tags.get("DeploymentId")
        if not deployment_id:
            self._log.error("Deployment ID not found in Public IP tags")
            return self._run_del_reqs(del_reqs)

        deployment_name = self.vm_deployment_name_tmpl.format(
            self.rg_name, deployment_id
        )
        try:
            deployment = self.resource_client.deployments.get(
                self.rg_name, deployment_name
            )
            req = DeleteRequest(
                99, self.resource_client.deployments, deployment_name
            )
            del_reqs.append(req)
        except Exception as e:
            self._log.error(
                f"Can't get deployment {deployment_name}: {str(e)}"
            )
            return self._run_del_reqs(del_reqs)

        deps = (
            cast(
                Union[List[Dependency], None],
                deployment.properties and deployment.properties.dependencies,
            )
            or []
        )
        for d in deps:
            if not d.resource_name:
                continue
            if d.resource_type == "Microsoft.Compute/virtualMachines":
                op = self.compute_client.virtual_machines
                del_reqs.append(DeleteRequest(0, op, d.resource_name))
            if d.resource_type == "Microsoft.Network/networkInterfaces":
                op = self.network_client.network_interfaces
                del_reqs.append(DeleteRequest(1, op, d.resource_name))
        return self._run_del_reqs(del_reqs)



#!/usr/bin/env python3

from os import getenv

__version__ = "0.9.0"

CONFIG_FILE = getenv(
    "YASCHEDULER_CONF_PATH", "/etc/yascheduler/yascheduler.conf"
)
LOG_FILE = getenv("YASCHEDULER_LOG_PATH", "/var/log/yascheduler.log")
PID_FILE = getenv("YASCHEDULER_PID_PATH", "/var/run/yascheduler.pid")
SLEEP_INTERVAL = 6
N_IDLE_PASSES = 20
DEFAULT_NODES_PER_PROVIDER = 10



#!/usr/bin/env python3

import logging
import queue
import random
from configparser import NoSectionError
from typing import List, Optional

from .abstract_cloud_api import load_cloudapi
from .workers import (
    AllocateResult,
    AllocateTask,
    AllocatorWorker,
    BackgroundWorker,
    DeallocateResult,
    DeallocateTask,
    DeallocatorWorker,
)
import yascheduler.scheduler

for logger_name in [
    "paramiko.transport",
]:
    logging.getLogger(logger_name).setLevel(logging.ERROR)


class CloudAPIManager(object):
    _log: logging.Logger
    _allocators: List[AllocatorWorker]
    _allocate_tasks: "queue.Queue[AllocateTask]"
    _allocate_results: "queue.Queue[AllocateResult]"
    _deallocators: List[DeallocatorWorker]
    _deallocate_tasks: "queue.Queue[DeallocateTask]"
    _deallocate_results: "queue.Queue[DeallocateResult]"
    yascheduler: "Optional['yascheduler.scheduler.Yascheduler']"

    def __init__(self, config, logger: Optional[logging.Logger] = None):
        if logger:
            self._log = logger.getChild(self.__class__.__name__)
        else:
            self._log = logging.getLogger(self.__class__.__name__)
        self.apis = {}
        self.yascheduler = None
        self.tasks = set()
        active_providers = set()

        try:
            for name, value in dict(config.items("clouds")).items():
                if not value:
                    continue
                active_providers.add(name.split("_")[0])
        except NoSectionError:
            pass

        for name in active_providers:
            if (
                config.getint("clouds", name + "_max_nodes", fallback=None)
                == 0
            ):
                continue
            self.apis[name] = load_cloudapi(name)(config)

        self._log.info(
            "Active cloud APIs: " + (", ".join(self.apis.keys()) or "-")
        )

        self._allocate_tasks = queue.Queue()
        self._allocate_results = queue.Queue()
        allocator_thread_num = int(
            config.get("local", "allocator_threads", fallback=10)
        )
        self._allocators = [
            AllocatorWorker(
                name=f"AllocatorThread[{x}]",
                logger=self._log,
                config=config,
                use_apis=self.apis.keys(),
                task_queue=self._allocate_tasks,
                result_queue=self._allocate_results,
            )
            for x in range(allocator_thread_num)
        ]

        self._deallocate_tasks = queue.Queue()
        self._deallocate_results = queue.Queue()
        deallocator_thread_num = int(
            config.get("local", "deallocator_threads", fallback=2)
        )
        self._deallocators = [
            DeallocatorWorker(
                name=f"DellocatorThread[{x}]",
                logger=self._log,
                config=config,
                use_apis=self.apis.keys(),
                task_queue=self._deallocate_tasks,
                result_queue=self._deallocate_results,
            )
            for x in range(deallocator_thread_num)
        ]

    def stop(self):
        self._log.info("Stopping threads...")
        workers: List[BackgroundWorker] = []
        workers.extend(self._allocators)
        workers.extend(self._deallocators)
        for t in workers:
            t.stop()
        for t in workers:
            t.join()
        self.do_async_work()

    def __bool__(self):
        return bool(len(self.apis))

    def initialize(self):
        assert self.yascheduler
        for cloudapi in self.apis:
            self.apis[cloudapi].yascheduler = self.yascheduler
            self.apis[cloudapi].init_key()

        for t in self._allocators:
            t.start()
        for t in self._deallocators:
            t.start()

    def allocate_node(self) -> str:
        assert self.yascheduler
        c = self.yascheduler.cursor
        active_providers = list(self.apis.keys())
        used_providers = []
        c.execute(
            """
            SELECT cloud, COUNT(cloud)
            FROM yascheduler_nodes
            WHERE cloud IS NOT NULL GROUP BY cloud;
            """
        )
        for row in c.fetchall():
            cloudapi = self.apis.get(row[0])
            if not cloudapi:
                continue
            if row[1] >= cloudapi.max_nodes:
                active_providers.remove(cloudapi.name)
                continue
            used_providers.append((row[0], row[1]))

        if not active_providers:
            self._log.warning("No suitable cloud provides")

        self._log.info("Enabled: %s" % str(active_providers))
        self._log.info("In use : %s" % str(used_providers))

        if len(used_providers) < len(active_providers):
            name = random.choice(
                list(
                    set(active_providers) - set([x[0] for x in used_providers])
                )
            )
        else:
            name = sorted(used_providers, key=lambda x: x[1])[0][0]

        cloudapi = self.apis[name]
        self._log.info("Chosen: %s" % cloudapi.name)

        c.execute(
            """INSERT INTO yascheduler_nodes (ip, enabled, cloud) VALUES (
            'prov' || SUBSTR(MD5(RANDOM()::TEXT), 0, 11),
            FALSE,
            %s
        ) RETURNING ip;""",
            [cloudapi.name],
        )

        t = AllocateTask(api_name=cloudapi.name, tmp_ip=c.fetchone()[0])
        self._allocate_tasks.put(t)
        self.yascheduler.connection.commit()
        return t.tmp_ip

    def allocate(self, on_task):
        if on_task in self.tasks:
            return
        self.tasks.add(on_task)
        self.allocate_node()

    def process_allocated(self):
        assert self.yascheduler
        c = self.yascheduler.cursor
        while not self._allocate_results.empty():
            try:
                r = self._allocate_results.get(False)
            except queue.Empty:
                break

            c.execute("DELETE FROM yascheduler_nodes WHERE ip=%s;", [r.tmp_ip])
            if r.ip and r.provisioned:
                c.execute(
                    """
                    INSERT INTO yascheduler_nodes (ip, ncpus, cloud)
                    VALUES (%s, %s, %s);
                    """,
                    [r.ip, r.ncpus, r.api_name],
                )
            if r.ip and not r.provisioned:
                self.deallocate([r.ip])

            self.yascheduler.connection.commit()
            self._allocate_results.task_done()

    def deallocate(self, ips):
        assert self.yascheduler
        c = self.yascheduler.cursor
        c.execute(
            "UPDATE yascheduler_nodes SET enabled=false WHERE ip IN ('%s');"
            % "', '".join(ips)
        )
        self.yascheduler.connection.commit()
        c.execute(
            """
            SELECT ip, cloud
            FROM yascheduler_nodes
            WHERE cloud IS NOT NULL AND ip IN ('%s');
            """
            % "', '".join(ips)
        )
        for row in c.fetchall():
            t = DeallocateTask(ip=row[0], api_name=row[1])
            self._deallocate_tasks.put(t)

    def process_deallocated(self):
        assert self.yascheduler
        c = self.yascheduler.cursor
        while not self._deallocate_results.empty():
            try:
                r = self._deallocate_results.get(False)
            except queue.Empty:
                break

            c.execute("DELETE FROM yascheduler_nodes WHERE ip=%s;", [r.ip])
            self.yascheduler.connection.commit()
            self._deallocate_results.task_done()

    def do_async_work(self):
        self.process_allocated()
        self.process_deallocated()

    def get_capacity(self, resources):
        n_busy_cloud_nodes = len(
            [item for item in resources if item[3]]
        )  # Yascheduler.queue_get_resources()
        max_nodes = sum(
            [self.apis[cloudapi].max_nodes for cloudapi in self.apis]
        )
        diff = max_nodes - n_busy_cloud_nodes
        return diff if diff > 0 else 0



#!/usr/bin/env python3

import inspect
import logging
import os
import random
import string

from configparser import ConfigParser
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import (
    default_backend as crypto_default_backend,
)
from datetime import timedelta, datetime
from importlib import import_module
from time import sleep
from typing import Any, Callable, Dict, Optional, TypeVar

from fabric import Connection as SSH_Connection
from paramiko.rsakey import RSAKey
import yascheduler.scheduler
from yascheduler import DEFAULT_NODES_PER_PROVIDER

T = TypeVar("T")


class AbstractCloudAPI(object):

    _log: logging.Logger
    name: str = "abstract"
    config: ConfigParser
    yascheduler: "Optional['yascheduler.scheduler.Yascheduler']"

    def __init__(
        self,
        max_nodes: Optional[int] = None,
        logger: Optional[logging.Logger] = None,
    ):
        if logger:
            self.log = logger.getChild(self.name)
        else:
            self._log = logging.getLogger(self.name)
        self.max_nodes = int(
            max_nodes if max_nodes is not None else DEFAULT_NODES_PER_PROVIDER
        )
        self.yascheduler = None

        self.public_key = None
        self.ssh_custom_key = None

    @property
    def ssh_user(self) -> str:
        "Default SSH user"
        return self.config.get(
            "clouds",
            f"{self.name}_user",
            fallback=self.config.get("remote", "user", fallback="root"),
        )

    def init_key(self):
        if self.ssh_custom_key:
            return

        for filename in os.listdir(self.config.get("local", "data_dir")):
            if not filename.startswith("yakey") or not os.path.isfile(
                os.path.join(self.config.get("local", "data_dir"), filename)
            ):
                continue
            key_path = os.path.join(
                self.config.get("local", "data_dir"), filename
            )
            self.key_name = key_path.split(os.sep)[-1]
            pmk_key = RSAKey.from_private_key_file(key_path)
            self._log.info("LOADED KEY %s" % key_path)
            break

        else:
            self.key_name = self.get_rnd_name("yakey")
            key = rsa.generate_private_key(
                backend=crypto_default_backend(),
                public_exponent=65537,
                key_size=2048,
            )
            pmk_key = RSAKey(key=key)
            key_path = os.path.join(
                self.config.get("local", "data_dir"), self.key_name
            )
            pmk_key.write_private_key_file(key_path)
            self._log.info("WRITTEN KEY %s" % key_path)

        self.public_key = "%s %s" % (pmk_key.get_name(), pmk_key.get_base64())

        self.ssh_custom_key = {"pkey": pmk_key}
        if self.yascheduler:
            self.yascheduler.ssh_custom_key = self.ssh_custom_key

    def get_rnd_name(self, prefix):
        return (
            prefix
            + "-"
            + "".join(
                [random.choice(string.ascii_lowercase) for _ in range(8)]
            )
        )

    @property
    def cloud_config_data(self) -> Dict[str, Any]:
        "Common cloud-config"
        return {
            "package_upgrade": True,
            "packages": ["openmpi-bin"],
        }

    def _retry_with_backoff(
        self,
        fn: Callable[[], T],
        max_time: float = 60,
        max_interval: float = 10,
    ) -> T:
        "Retry with random backoff"
        end_time = datetime.now() + timedelta(seconds=max_time)
        while True:
            try:
                return fn()
            except Exception as e:
                if datetime.now() >= end_time:
                    raise e
                sleep(
                    min(
                        random.random() * max_interval,
                        max(0, (datetime.now() - end_time).total_seconds()),
                    )
                )

    def _run_ssh_cmd_with_backoff(
        self,
        host: str,
        cmd: str,
        max_time: float = 60,
        max_interval: float = 10,
    ):
        "Run ssh command with retries on errors"

        def run_cmd():
            ssh_conn = SSH_Connection(
                host=host,
                user=self.ssh_user,
                connect_kwargs=self.ssh_custom_key,
            )
            return ssh_conn.run(cmd, hide=True)

        return self._retry_with_backoff(run_cmd, max_time, max_interval)

    def create_node(self) -> str:
        raise NotImplementedError()

    def setup_node(self, ip):
        ssh_conn = SSH_Connection(
            host=ip, user=self.ssh_user, connect_kwargs=self.ssh_custom_key
        )
        sudo_prefix = "" if self.ssh_user == "root" else "sudo "
        apt_cmd = f"{sudo_prefix}apt-get -o DPkg::Lock::Timeout=600"
        ssh_conn.run(f"{apt_cmd} -y update && {apt_cmd} -y upgrade", hide=True)
        ssh_conn.run(f"{apt_cmd} -y install openmpi-bin", hide=True)

        ssh_conn.run("mkdir -p ~/bin", hide=True)
        if self.config.get("local", "deployable").startswith("http"):
            # downloading binary from a trusted non-public address
            ssh_conn.run(
                "cd ~/bin && wget %s" % self.config.get("local", "deployable"),
                hide=True,
            )
        else:
            # uploading binary from local; requires broadband connection
            # ssh_conn.put(
            #     self.config.get('local', 'deployable'), 'bin/dummyengine'
            # ) # TODO
            ssh_conn.put(
                self.config.get("local", "deployable"), "bin/Pcrystal"
            )  # TODO

        if self.config.get("local", "deployable").endswith(".gz"):
            # binary may be gzipped, without subfolders, with an arbitrary
            # archive name, but the name of the binary must remain Pcrystal
            ssh_conn.run(
                "cd ~/bin && tar xvf %s"
                % self.config.get("local", "deployable").split("/")[-1],
                hide=True,
            )
        # ssh_conn.run('ln -sf ~/bin/Pcrystal /usr/bin/Pcrystal', hide=True)

        # print and ensure versions
        result = ssh_conn.run(
            "/usr/bin/mpirun --allow-run-as-root -V", hide=True
        )
        self._log.info(result.stdout)
        result = ssh_conn.run("cat /etc/issue", hide=True)
        self._log.info(result.stdout)
        result = ssh_conn.run("grep -c ^processor /proc/cpuinfo", hide=True)
        self._log.info(result.stdout)

    def delete_node(self, ip: str):
        raise NotImplementedError()


def load_cloudapi(name):
    cloudapi_mod = import_module("." + name, package="yascheduler.clouds")
    for _, cls in inspect.getmembers(cloudapi_mod):
        if inspect.isclass(cls) and cls.__base__ == AbstractCloudAPI:
            return cls
    raise ImportError



"""
Aiida plugin for yascheduler
"""

import aiida.schedulers
from aiida.schedulers.datastructures import (JobState, JobInfo, NodeNumberJobResource)

_MAP_STATUS_YASCHEDULER = {
    'QUEUED': JobState.QUEUED,
    'RUNNING': JobState.RUNNING,
    'FINISHED': JobState.DONE
}


class YaschedJobResource(NodeNumberJobResource):

    def __init__(self, *_, **kwargs):
        super(YaschedJobResource, self).__init__(**kwargs)


class YaScheduler(aiida.schedulers.Scheduler):
    """
    Support for the YaScheduler designed specifically for MPDS
    """
    _logger = aiida.schedulers.Scheduler._logger.getChild('yascheduler')

    # Query only by list of jobs and not by user
    _features = {
        'can_query_by_user': False,
    }

    # The class to be used for the job resource.
    _job_resource_class = YaschedJobResource

    def _get_joblist_command(self, jobs=None, user=None):
        """
        The command to report full information on existing jobs.
        """
        from aiida.common.exceptions import FeatureNotAvailable
        if user:
            raise FeatureNotAvailable("Cannot query by user in Yascheduler")
        command = ["yastatus"]
        # make list from job ids (taken from slurm scheduler)
        if jobs:
            joblist = []
            if isinstance(jobs, str):
                joblist.append(jobs)
            else:
                if not isinstance(jobs, (tuple, list)):
                    raise TypeError("If provided, the 'jobs' variable must be a string or a list of strings")
                joblist = jobs
            command.append('--jobs {}'.format(' '.join(joblist)))
        return ' '.join(command)

    def _get_detailed_jobinfo_command(self, jobid):
        """
        Return the command to run to get the detailed information on a job,
        even after the job has finished.
        """
        return 'yastatus --jobs {}'.format(jobid)

    def _get_submit_script_header(self, job_tmpl):
        """
        Return the submit script header, using the parameters from the
        job_tmpl.
        """
        lines = []
        if job_tmpl.job_name:
            lines.append("LABEL={}".format(job_tmpl.job_name))

        # TODO too specific for engine.pcrystal
        lines += ["INPUT=INPUT", "STRUCT=fort.34"]
        return "\n".join(lines)

    def _get_submit_command(self, submit_script):
        """
        Return the string to execute to submit a given script.
        """
        return 'yasubmit {}'.format(submit_script)

    def _parse_submit_output(self, retval, stdout, stderr):
        """
        Parse the output of the submit command, as returned by executing the
        command returned by _get_submit_command command.
        """
        if stderr.strip():
            self.logger.warning("Stderr when submitting: {}".format(stderr.strip()))
        return stdout.split(':')[1].strip()

    def _parse_joblist_output(self, retval, stdout, stderr):
        """
        Parse the queue output string, as returned by executing the
        command returned by _get_joblist_command command,
        that is here implemented as a list of lines, one for each
        job, with _field_separator as separator. The order is described
        in the _get_joblist_command function.

        Return a list of JobInfo objects, one of each job,
        each relevant parameters implemented.
        """
        if stderr.strip():
            self.logger.warning("Stderr when parsing joblist: {}".format(stderr.strip()))
        job_list = [job.split() for job in stdout.split('\n') if job]
        job_infos = []
        for job_id, status in job_list:
            job = JobInfo()
            job.job_id = job_id
            job.job_state = _MAP_STATUS_YASCHEDULER[status]
            job_infos.append(job)
        return job_infos

    def _get_kill_command(self, jobid):
        """
        Return the command to kill the job with specified jobid.
        """

    def _parse_kill_output(self, retval, stdout, stderr):
        """
        Parse the output of the kill command.
        """



import json

from hcloud import Client, APIException
from hcloud.images.domain import Image
from hcloud.server_types.domain import ServerType
from hcloud.ssh_keys.domain import SSHKey

from yascheduler.clouds import AbstractCloudAPI


class HetznerCloudAPI(AbstractCloudAPI):

    name = 'hetzner'

    def __init__(self, config):
        super().__init__(max_nodes=config.getint('clouds', 'hetzner_max_nodes', fallback=None))
        self.client = Client(token=config.get('clouds', 'hetzner_token'))
        self.config = config

    def init_key(self):
        super().init_key()
        try:
            self.ssh_key_id = self.client.ssh_keys.create(name=self.key_name, public_key=self.public_key)
        except APIException as ex:
            if 'already' in str(ex):
                for key in self.client.ssh_keys.get_all():
                    if key.name.startswith('yakey') and len(key.name) == 14:
                        self.ssh_key_id = key.id
            else:
                raise

    def create_node(self):
        assert self.ssh_key_id
        assert self.ssh_custom_key

        response = self.client.servers.create(
            name=self.get_rnd_name('node'),
            server_type=ServerType('cx51'), image=Image(name='debian-10'),
            ssh_keys=[SSHKey(name=self.key_name)],
            user_data="#cloud-config\n" + json.dumps(self.cloud_config_data),
        )
        server = response.server
        ip = server.public_net.ipv4.ip
        self._log.info('CREATED %s' % ip)

        # wait node up and ready
        self._run_ssh_cmd_with_backoff(
            ip, cmd="cloud-init status --wait", max_interval=5
        )

        return ip

    def delete_key(self):
        self.client.ssh_keys.delete(self.ssh_key_id)

    def delete_node(self, ip):
        server = None

        for s in self.client.servers.get_all():
            if s.public_net.ipv4.ip == ip:
                server = self.client.servers.get_by_id(s.id)
                break

        if server:
            server.delete()
            self._log.info('DELETED %s' % ip)

        else:
            self._log.info('NODE %s NOT DELETED AS UNKNOWN' % ip)
